#include <windows.h>
#include <stdio.h>
// Safer versions
#include <strsafe.h>

// Function prototypes
// A prototype that receives a function name, displaying
// system error code and its respective message
void DisplayErrorBox(LPTSTR lpszFunction);
void RefreshDirectory(LPTSTR);
void RefreshTree(LPTSTR);
void WatchDirectory(LPTSTR);

int wmain(int argc, WCHAR *argv[])
{
    if(argc != 2)
    {
		wprintf(L"Change notification for directory\n");
        wprintf(L"Usage: %s <dir>\n", argv[0]);
        return 1;
    }

    WatchDirectory(argv[1]);
	return 0;
}

void DisplayErrorBox(LPTSTR lpszFunction) 
{ 
    // Retrieve the system error message for the last-error code
    LPVOID lpMsgBuf;
    LPVOID lpDisplayBuf;
    DWORD dw = GetLastError(); 

    FormatMessage(
        FORMAT_MESSAGE_ALLOCATE_BUFFER | 
        FORMAT_MESSAGE_FROM_SYSTEM |
        FORMAT_MESSAGE_IGNORE_INSERTS,
        NULL,
        dw,
        MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT),
        (LPTSTR) &lpMsgBuf,
        0, NULL );

    // Display the error message and clean up
    lpDisplayBuf = (LPVOID)LocalAlloc(LMEM_ZEROINIT, (lstrlen((LPCTSTR)lpMsgBuf)+lstrlen((LPCTSTR)lpszFunction)+40)*sizeof(WCHAR)); 
    StringCchPrintf((LPTSTR)lpDisplayBuf, LocalSize(lpDisplayBuf) / sizeof(WCHAR),  L"%s failed with error %d: %s", 
        lpszFunction, dw, lpMsgBuf); 
    MessageBox(NULL, (LPCTSTR)lpDisplayBuf, L"Error", MB_OK); 

    LocalFree(lpMsgBuf);
    LocalFree(lpDisplayBuf);
}

void WatchDirectory(LPTSTR lpDir)
{
   DWORD dwWaitStatus; 
   HANDLE dwChangeHandles[2]; 
   WCHAR lpDrive[4];
   WCHAR lpFile[_MAX_FNAME];
   WCHAR lpExt[_MAX_EXT];

   // Breaks a path name into components. These are versions of _splitpath(),
   // _wsplitpath() with security enhancements
   _wsplitpath_s(lpDir, lpDrive, 4, NULL, 0, lpFile, _MAX_FNAME, lpExt, _MAX_EXT);

   lpDrive[2] = '\\';
   lpDrive[3] = '\0';
   
   // Watch the directory for file creation and deletion 
   dwChangeHandles[0] = FindFirstChangeNotification( 
      lpDir,                         // directory to watch 
      FALSE,                         // do not watch subtree 
      FILE_NOTIFY_CHANGE_FILE_NAME); // watch file name changes 
 
   if (dwChangeHandles[0] == INVALID_HANDLE_VALUE) 
   {
     DisplayErrorBox(L"FindFirstChangeNotification() 1");
     ExitProcess(GetLastError()); 
   }
   
   // Watch the subtree for directory creation and deletion 
   dwChangeHandles[1] = FindFirstChangeNotification( 
      lpDrive,                       // directory to watch 
      TRUE,                          // watch the subtree 
      FILE_NOTIFY_CHANGE_DIR_NAME);  // watch dir name changes 
 
   if (dwChangeHandles[1] == INVALID_HANDLE_VALUE) 
   {
     DisplayErrorBox(L"FindFirstChangeNotification() 2");
     ExitProcess(GetLastError()); 
   }
   
   // Make a final validation check on our handles
   if ((dwChangeHandles[0] == NULL) || (dwChangeHandles[1] == NULL))
   {
     wprintf(L"\nFindFirstChangeNotification() - unexpected NULL.\n");
     ExitProcess(GetLastError()); 
   }
   
   // Change notification is set. Now wait on both notification 
   // handles and refresh accordingly 
   while (TRUE) 
   {
	   // Wait for notification 
	   wprintf(L"\nWaiting for notification...\n");

      dwWaitStatus = WaitForMultipleObjects(2, dwChangeHandles, FALSE, INFINITE); 
 
      switch (dwWaitStatus) 
      { 
         case WAIT_OBJECT_0:			 
			 // A file was created, renamed, or deleted in the directory.
			 // Refresh this directory and restart the notification 
             RefreshDirectory(lpDir); 
             if (FindNextChangeNotification(dwChangeHandles[0]) == FALSE )
             {
               DisplayErrorBox(L"FindNextChangeNotification()");
               ExitProcess(GetLastError()); 
             }
             break; 
 
         case WAIT_OBJECT_0 + 1:
			 // A directory was created, renamed, or deleted.
			 // Refresh the tree and restart the notification 
             RefreshTree(lpDrive); 
             if (FindNextChangeNotification(dwChangeHandles[1]) == FALSE )
             {
               DisplayErrorBox(L"FindNextChangeNotification()");
               ExitProcess(GetLastError()); 
             }
             break; 
 
         case WAIT_TIMEOUT:
			 // A timeout occurred, this would happen if some value other 
			 // than INFINITE is used in the Wait call and no changes occur.
			 // In a single-threaded environment you might not want an INFINITE wait 
            wprintf(L"\nNo changes in the timeout period.\n");
            break;

         default: 
            wprintf(L"\n ERROR: Unhandled dwWaitStatus.\n");
            ExitProcess(GetLastError());
            break;
      }
   }
}

void RefreshDirectory(LPTSTR lpDir)
{
   // This is where you might place code to refresh your
   // directory listing, but not the subtree because it
   // would not be necessary
   wprintf(L"Directory %s changed!\n", lpDir);
}

void RefreshTree(LPTSTR lpDrive)
{
   // This is where you might place code to refresh your
   // directory listing, including the subtree.
   wprintf(L"Directory tree %s changed!\n", lpDrive);
}

